# coding=utf-8
from __future__ import absolute_import, division, print_function, unicode_literals

import wrapt
from flask_sqlalchemy import SQLAlchemy

import scout_apm.sqlalchemy


def instrument_sqlalchemy(db):
    SQLAlchemy.get_engine = wrapped_get_engine(SQLAlchemy.get_engine)


@wrapt.decorator
def wrapped_get_engine(wrapped, instance, args, kwargs):
    engine = wrapped(*args, **kwargs)
    scout_apm.sqlalchemy.instrument_sqlalchemy(engine)
    return engine



# coding=utf-8
from __future__ import absolute_import, division, print_function, unicode_literals

import logging

import wrapt

from scout_apm.compat import text_type
from scout_apm.core.config import scout_config
from scout_apm.core.tracked_request import TrackedRequest

try:
    from urllib3 import HTTPConnectionPool
except ImportError:  # pragma: no cover
    HTTPConnectionPool = None

logger = logging.getLogger(__name__)

have_patched_pool_urlopen = False


def ensure_installed():
    global have_patched_pool_urlopen

    logger.debug("Instrumenting urllib3.")

    if HTTPConnectionPool is None:
        logger.debug(
            "Couldn't import urllib3.HTTPConnectionPool - probably not installed."
        )
        return False
    elif not have_patched_pool_urlopen:
        try:
            HTTPConnectionPool.urlopen = wrapped_urlopen(HTTPConnectionPool.urlopen)
        except Exception as exc:
            logger.warning(
                "Failed to instrument for Urllib3 HTTPConnectionPool.urlopen: %r",
                exc,
                exc_info=exc,
            )
        else:
            have_patched_pool_urlopen = True


@wrapt.decorator
def wrapped_urlopen(wrapped, instance, args, kwargs):
    def _extract_method(method, *args, **kwargs):
        return method

    try:
        method = _extract_method(*args, **kwargs)
    except TypeError:
        method = "Unknown"

    try:
        url = text_type(instance._absolute_url("/"))
    except Exception:
        logger.exception("Could not get URL for HTTPConnectionPool")
        url = "Unknown"

    # Don't instrument ErrorMonitor calls
    if text_type(url).startswith(scout_config.value("errors_host")):
        return wrapped(*args, **kwargs)

    tracked_request = TrackedRequest.instance()
    with tracked_request.span(operation="HTTP/{}".format(method)) as span:
        span.tag("url", text_type(url))
        return wrapped(*args, **kwargs)



# coding=utf-8
from __future__ import absolute_import, division, print_function, unicode_literals

from nameko.extensions import DependencyProvider
from nameko.web.handlers import HttpRequestHandler
from werkzeug.wrappers import Request, Response

import scout_apm.core
from scout_apm.core.tracked_request import TrackedRequest
from scout_apm.core.web_requests import werkzeug_track_request_data


class ScoutReporter(DependencyProvider):
    def setup(self):
        installed = scout_apm.core.install()
        self._do_nothing = not installed

    def worker_setup(self, worker_ctx):
        if self._do_nothing:
            return
        tracked_request = TrackedRequest.instance()
        tracked_request.is_real_request = True

        # Get HTTP details for HTTP handlers
        if isinstance(worker_ctx.entrypoint, HttpRequestHandler):
            try:
                request = worker_ctx.args[0]
            except IndexError:
                pass
            else:
                if isinstance(request, Request):
                    werkzeug_track_request_data(request, tracked_request)

        operation = (
            "Controller/"
            + worker_ctx.service.name
            + "."
            + worker_ctx.entrypoint.method_name
        )
        tracked_request.start_span(operation=operation, should_capture_backtrace=False)

    def worker_result(self, worker_ctx, result=None, exc_info=None):
        if self._do_nothing:
            return
        tracked_request = TrackedRequest.instance()

        if exc_info:
            tracked_request.tag("error", "true")
        elif isinstance(worker_ctx.entrypoint, HttpRequestHandler):
            # Handle the cases that HttpRequestHandler.response_from_result
            # does
            if isinstance(result, Response):
                status_code = result.status_code
            elif isinstance(result, tuple):
                if len(result) == 3:
                    status_code, _headers, _payload = result
                elif len(result) == 2:
                    status_code, _payload = result
                else:
                    # Nameko doesn't support other formats, so we know it will
                    # turn this into an error
                    status_code = 500
            else:
                status_code = 200

            if 500 <= status_code <= 599:
                tracked_request.tag("error", "true")

        tracked_request.stop_span()
